package com.revolsys.gis.data.model;

import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import com.revolsys.gis.data.model.types.DataType;
import com.revolsys.io.AbstractObjectWithProperties;

/**
 * The Attribute class defines the name, type and other properties about each
 * attribute on a {@link DataObject} in the {@link DataObjectMetaData}.
 * 
 * @author Paul Austin
 * @see DataObject
 * @see DataObjectMetaData
 */
public class Attribute extends AbstractObjectWithProperties implements
  Cloneable {
  private final Map<Object, Object> allowedValues = new LinkedHashMap<Object, Object>();

  private Object defaultValue;

  /** The description of the attribute. */
  private String description;

  private int index;

  /** The maximum length of an attribute value. */
  private int length;

  /** The name of the attribute. */
  private String name;

  /** The flag indicating if a value is required for the attribute. */
  private boolean required;

  /** The maximum number of decimal places. */
  private int scale;

  /** The data type of the attribute value. */
  private DataType type;

  public Attribute() {
  }

  public Attribute(final Attribute attribute) {
    this.name = attribute.getName();
    this.description = attribute.getDescription();
    this.type = attribute.getType();
    this.required = attribute.isRequired();
    this.length = attribute.getLength();
    this.scale = attribute.getScale();
    final Map<String, Object> properties = attribute.getProperties();
    setProperties(properties);
  }

  public Attribute(final int index) {
    this.index = index;
  }

  /**
   * Construct a new attribute.
   * 
   * @param name The name of the attribute.
   * @param type The data type of the attribute value.
   * @param required The flag indicating if a value is required for the
   *          attribute.
   */
  public Attribute(final String name, final DataType type,
    final boolean required) {
    this(name, type, 0, 0, required, null, null);
  }

  /**
   * Construct a new attribute.
   * 
   * @param name The name of the attribute.
   * @param type The data type of the attribute value.
   * @param length The maximum length of an attribute value, 0 for no maximum.
   * @param required The flag indicating if a value is required for the
   *          attribute.
   * @param properties The meta data properties about the attribute.
   */
  public Attribute(final String name, final DataType type,
    final boolean required, final Map<String, Object> properties) {
    this(name, type, 0, 0, required, properties);
  }

  public Attribute(final String name, final DataType dataType,
    final boolean required, final String description) {
    this(name, dataType, 0, 0, required, description, null);
  }

  /**
   * Construct a new attribute.
   * 
   * @param name The name of the attribute.
   * @param type The data type of the attribute value.
   * @param length The maximum length of an attribute value, 0 for no maximum.
   * @param required The flag indicating if a value is required for the
   *          attribute.
   */
  public Attribute(final String name, final DataType type, final int length,
    final boolean required) {
    this(name, type, length, 0, required, null, null);
  }

  /**
   * Construct a new attribute.
   * 
   * @param name The name of the attribute.
   * @param type The data type of the attribute value.
   * @param length The maximum length of an attribute value, 0 for no maximum.
   * @param scale The maximum number of decimal places.
   * @param required The flag indicating if a value is required for the
   *          attribute.
   * @param properties The meta data properties about the attribute.
   */
  public Attribute(final String name, final DataType type, final int length,
    final boolean required, final Map<String, Object> properties) {
    this(name, type, length, 0, required, properties);
  }

  public Attribute(final String name, final DataType type, final int length,
    final boolean required, final String description) {
    this(name, type, length, 0, required, description, null);
  }

  /**
   * Construct a new attribute.
   * 
   * @param name The name of the attribute.
   * @param type The data type of the attribute value.
   * @param required The flag indicating if a value is required for the
   *          attribute.
   * @param properties The meta data properties about the attribute.
   */
  public Attribute(final String name, final DataType type,
    final Integer length, final Integer scale, final Boolean required) {
    this(name, type, length, scale, required, "");
  }

  /**
   * Construct a new attribute.
   * 
   * @param name The name of the attribute.
   * @param type The data type of the attribute value.
   * @param length The maximum length of an attribute value, 0 for no maximum.
   * @param scale The maximum number of decimal places.
   * @param required The flag indicating if a value is required for the
   *          attribute.
   * @param properties The meta data properties about the attribute.
   */
  public Attribute(final String name, final DataType type,
    final Integer length, final Integer scale, final Boolean required,
    final Map<String, Object> properties) {
    this(name, type, length, scale, required, null, properties);

  }

  /**
   * Construct a new attribute.
   * 
   * @param name The name of the attribute.
   * @param type The data type of the attribute value.
   * @param length The maximum length of an attribute value, 0 for no maximum.
   * @param scale The maximum number of decimal places.
   * @param required The flag indicating if a value is required for the
   *          attribute.
   * @param properties The meta data properties about the attribute.
   */
  public Attribute(final String name, final DataType type,
    final Integer length, final Integer scale, final Boolean required,
    final String description) {
    this.name = name;
    this.description = description;
    this.type = type;
    if (required != null) {
      this.required = required;
    }
    if (length != null) {
      this.length = length;
    }
    if (scale != null) {
      this.scale = scale;
    }
    this.description = description;
  }

  /**
   * Construct a new attribute.
   * 
   * @param name The name of the attribute.
   * @param type The data type of the attribute value.
   * @param length The maximum length of an attribute value, 0 for no maximum.
   * @param scale The maximum number of decimal places.
   * @param required The flag indicating if a value is required for the
   *          attribute.
   * @param properties The meta data properties about the attribute.
   */
  public Attribute(final String name, final DataType type,
    final Integer length, final Integer scale, final Boolean required,
    final String description, final Map<String, Object> properties) {
    this.name = name;
    this.type = type;
    if (required != null) {
      this.required = required;
    }
    if (length != null) {
      this.length = length;
    }
    if (scale != null) {
      this.scale = scale;
    }
    this.description = description;
    setProperties(properties);
  }

  public void addAllowedValue(final Object value, final Object text) {
    this.allowedValues.put(value, text);
  }

  @Override
  public Attribute clone() {
    return new Attribute(this);
  }

  @Override
  public boolean equals(final Object object) {
    if (object instanceof Attribute) {
      final Attribute attribute = (Attribute)object;
      return (name.equals(attribute.getName()));
    } else {
      return false;
    }
  }

  public Map<Object, Object> getAllowedValues() {
    return allowedValues;
  }

  @SuppressWarnings("unchecked")
  public <T> T getDefaultValue() {
    return (T)defaultValue;
  }

  public String getDescription() {
    return description;
  }

  public int getIndex() {
    return index;
  }

  /**
   * Get the maximum length of the attribute value. The length 0 should be used
   * if there is no maximum.
   * 
   * @return The maximum length of an attribute value.
   */
  public int getLength() {
    return length;
  }

  public int getMaxStringLength() {
    int length = this.length;
    if (scale > 0) {
      length += 1;
      length += scale;
    }
    if (Number.class.isAssignableFrom(type.getJavaClass())) {
      length += 1;
    }
    return length;
  }

  /**
   * Get the name of the attribute.
   * 
   * @return The name of the attribute.
   */
  public String getName() {
    return name;
  }

  /**
   * Get the maximum number of decimal places of the attribute value.
   * 
   * @return The maximum number of decimal places.
   */
  public int getScale() {
    return scale;
  }

  /**
   * Get the data type of the attribute value.
   * 
   * @return The data type of the attribute value.
   */
  public DataType getType() {
    return type;
  }

  /**
   * Get the data type of the attribute value.
   * 
   * @return The data type of the attribute value.
   */
  public String getTypeDescription() {
    final StringBuffer typeDescription = new StringBuffer();
    typeDescription.append(type);
    if (length > 0) {
      typeDescription.append('(');
      typeDescription.append(length);
      if (scale > 0) {
        typeDescription.append(',');
        typeDescription.append(scale);
      }
      typeDescription.append(')');
    }
    return typeDescription.toString();
  }

  /**
   * Return the hash code of the attribute.
   * 
   * @return The hash code.
   */
  @Override
  public int hashCode() {
    return name.hashCode();
  }

  /**
   * Get the flag indicating if a value is required for the attribute.
   * 
   * @return True if a value is required, false otherwise.
   */
  public boolean isRequired() {
    return required;
  }

  public void setAllowedValues(final List<?> allowedValues) {
    for (final Object allowedValue : allowedValues) {
      this.allowedValues.put(allowedValue, allowedValue);
    }
  }

  public void setAllowedValues(final Map<?, ?> allowedValues) {
    this.allowedValues.putAll(allowedValues);
  }

  public void setDefaultValue(final Object defaultValue) {
    this.defaultValue = defaultValue;
  }

  public void setDescription(final String description) {
    this.description = description;
  }

  void setIndex(final int index) {
    this.index = index;
  }

  public void setRequired(final boolean required) {
    this.required = required;
  }

  public void setType(final DataType type) {
    this.type = type;
  }

  @Override
  public String toString() {
    final StringBuffer string = new StringBuffer(name);
    string.append(':');
    string.append(type);
    if (length > 0) {
      string.append('(');
      string.append(length);
      if (scale > 0) {
        string.append(',');
        string.append(scale);
      }
      string.append(')');
    }
    return string.toString();
  }
}
